#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <regex.h>

#define SHOKI_ALL	"ALL"
#define SHOKI_IGNORE	"IGNORE"

#define DOCTRINE_FILTER		1
#define DOCTRINE_FILE		2
#define DOCTRINE_COMMAND	3

#define DO_DURATION	1
#define DO_LABEL	2
#define DO_LOGALL	3
#define DO_REWRITE	4
#define DO_FILTER	5
#define DO_FILE		6
#define DO_COMMAND	7
#define DO_INSTANCES	8
#define DO_RESET	9

#define IP_S_HL         1
#define IP_S_V          2
#define IP_S_TOS        3
#define IP_S_LEN        4
#define IP_S_ID         5
#define IP_S_OFF        6
#define IP_S_TTL        7
#define IP_S_P          8
#define IP_S_SUM        9
#define IP_S_SRC        10
#define IP_S_DST        11

struct packet
{
	struct ip *ip;
	struct tcphdr *tcp;
	struct udphdr *udp;
	struct icmp *icmp;
	u_char *payload;
};

struct dconfig
{
	char label[MAX_CALLARG_LEN];
	char outfile[MAX_FILE_LEN];
	int items;				/* Rules in this doctrine */
	int lastmatch;				/* Last index matched */
	int matches;				/* Number of matches */
	int duration;				/* Window, this doctrine */
	int sec;				/* Start of window, sec */
	int usec;				/* Start of window, usec */
	int type;				/* Doctrine type */
	int log_all;				/* Log individual rules */
	int save_packets;
	int max_instances;
	pcap_t *pd;
	bpf_u_int32 netmask;
	struct doctrine *doctrine;		/* Linked list of rules */
	struct dconfig *next;
};

struct doctrine
{
	char filter[MAX_FILTER_LEN];		/* filter expression */
	char hex[MAX_HEX_LEN];			/* search expression */
	char call_args[MAX_CALLARG_LEN];	/* args */
	char id[MAX_ID_LEN];			/* unique filter id */
	char type[MAX_TYPE_LEN];		/* filter type */
	regex_t *regex;				/* compiled regex */
	struct bpf_program *fcode;		/* compiled fcode */
	int fcount;				/* snaplen */
	int d;					/* threshold duration */
	int t;					/* threshold value */
	int q;					/* threshold `quiet' time */
	int h;					/* threshold host count */
	int sev;				/* filter severity */
	int callback;				/* callback id */
	int dtype;				/* doctrine type */
	int matched;
	int index;
	int recompile;				/* recompile filter at run */
	struct pcap_pkthdr *hh;			/* pcap header */
	u_char *sp;				/* packet data */
	struct packet *packet;
	struct doctrine *next;
};

void parse_doctrine_config(const char *);
void parse_doctrine_filter(const char *, const char *, int, struct doctrine *, struct dconfig *);
void parse_doctrine_file(const char *, const char *, int, struct doctrine *, struct dconfig *);
void parse_doctrine_command(const char *, const char *, int, struct doctrine *, struct dconfig *);
void parse_doctrine_rule(const char *, const char *, int, struct doctrine *, struct dconfig *);
int handle_doctrine(const struct pcap_pkthdr *, const u_char *);
int check_tokens(const struct doctrine *, const char *, int);
char *get_bit(struct ip *, int);

int get_callback(char *);
