#!/usr/bin/perl 
#
# This is a little CGI frontend for many of sorts of routine queries.
# It requires that the perl CGI module be installed.
#
# The functionality of this script should be more or less identical
# to the functionality of the shoki_query script, which uses Tk.
#
# Here are some of the things that you'll probably want to tweak:
#
#	$dumpdir	If you run a query that will create a dump file,
#			it will go under this directory.
#	$tcpdump	A pointer to the tcpdump(1) binary, or similar
#			pcap dump file reader which outputs to stdout.
#			See the comments in the README about architecture-
#			specific issues if you're going to be creating
#			and looking at logs from multiple architectures.
#	$window		The default query window size, in seconds.
#	$mindumpsize	A minimum size, in bytes, for dump files.  Files
#			smaller than this won't be listed in the
#			log selection window.  This is particularly useful
#			when doing a filter match over a large window of
#			time.
#

use CGI;
use Time::Local;
&read_config("/etc/shoki.conf");
require("$INSTALL_PREFIX/$ETCDIR/collector.conf");
$logdir = "$INSTALL_PREFIX/central/log";
$dumpdir = "/usr/local/src/dump";
#$dumpdir = "/tmp";
$plotdir = "$INSTALL_PREFIX/plot";
@log_sensors = @sensors;
$tcpdump = '/usr/local/bin/tcpdump_fixed';
$window = 900;
$mindumpsize = 92;
$helpfile = "$INSTALL_PREFIX/$ETCDIR/help.txt";
$htmldir = "/home/httpd/html/plots";
$htmlref ="/plots";
@hidden_fields =
(
	'log_sensors',
	'selected_logs',
	'options',
	'filter',
	'Verbose',
);

&main();
exit(0);

sub main
{
	&do_topstuff();
	&check_query();
	&do_bottomstuff();
}


sub do_topstuff
{
	my($foo);
	$query = new CGI;
	foreach $foo ($query->param)
	{
		@{$foo} = $query->param($foo);
		@{$foo} = split('\s+', &check_filter(join(' ', @{$foo})));
	}
	print $query->header;
	print $query->start_html("Shoki Query Widget");
	print $query->startform;
}

sub do_bottomstuff
{
	my($foo);
	foreach $foo (@hidden_fields)
	{
		@{$foo} && (print $query->hidden($foo, @{$foo}));
	}
	print $query->endform;
	print $query->end_html;
}

sub check_query
{
	my(@val, $foo, $bar);
	if($query->param('action') eq 'Select Sensors')
	{
		&do_sensors();
	}
	elsif($query->param('action') eq 'Select Logs')
	{
		&do_logs();
	}
	elsif($query->param('action') eq 'Extract Logs')
	{
		&do_extract();
	}
	elsif($query->param('action') eq 'Extract')
	{
		&do_xq();
	}
	elsif($query->param('action') eq 'Dump')
	{
		&do_dump();
	}
	elsif($query->param('action') eq 'Sessions')
	{
		&do_widget('session');
	}
	elsif($query->param('action') eq 'Services')
	{
		&do_widget('service');
	}
	elsif($query->param('action') eq 'Plot')
	{
		&do_plot();
	}
	elsif($query->param('action') eq 'Print Values')
	{
		&do_foo();
	}
	else
	{
		&do_main();
	}
}

sub read_config
{
	my($fname) = @_;
	open(FILE, "<$fname") ||
		(die "Couldn't read $fname");
	while(<FILE>)
	{
		/^(\w+)(\s*)=(\s*)(.+)$/ || next;
		${$1} = $4;
	}
	close(FILE);
}

sub do_main
{
	print "<h1 align=center>Shoki Query Widget</h1><p>\n";
	print $query->submit('action', 'Select Sensors');
	print $query->submit('action', 'Select Logs');
	print $query->submit('action', 'Extract Logs');
	print $query->submit('action', 'Generic Query');
	print $query->submit('action', 'Print Values');
}

sub do_foo
{
	foreach $foo ($query->param)
	{
		@val = $query->param($foo);
		$bar = join(', ', @val);
		print "<p>$foo:  $bar\n"
	}
}
 
sub do_sensors
{
	print "<h1 align=center>Select one or more sensors</h1><p>\n";
	@log_sensors = ();
	print $query->scrolling_list(-name=>'log_sensors', -values=>\@sensors,
		-size=>5, -multiple=>'true');
	print "<br>\n";
	print $query->submit('action', 'Done');
}

sub do_logs
{
	my($host) = @_;
	my($foo, @bar, @baz);
	undef @selected_logs;
	undef @options;
	undef @filter;
	undef @Verbose;
	print "<h1 align=center>Select Log</h1><p>\n";
	$host || &get_logs();
	print "<b>Options: </b>";
	print $query->textfield('options');
	print "<br>\n";
	print "<b>Filters: </b>";
	print $query->textfield('filter');
	print "<br>\n";
	undef @baz;
	foreach $foo (sort(keys %logs))
	{
		#push(@bar, $dates{$foo});
		push(@baz, $foo);
	}
	print $query->scrolling_list(-name=>'selected_logs', -values=>\@baz,
		-labels=>\%dates, -size=>20, -multiple=>'true');
	print "<br>\n";
	print $query->submit('action', 'Dump');
	print $query->submit('action', 'Sessions');
	print $query->submit('action', 'Services');
	print $query->submit('action', 'Plot');
	print $query->checkbox('Verbose');
	print $query->submit('action', 'Back');
}

sub get_logs
{
	my($foo, $fname, $fname_w);
	undef %logs;
	undef %dates;
	foreach $foo (@log_sensors)
	{
		opendir(DIR, "$logdir/$foo") ||
			((print "<h3>Couldn't open dir $logdir/$foo</h3>"),
			return);
		while($fname = readdir(DIR))
		{
			($fname =~ /^$foo-sensor_dump\.(\d+)-(\d+)\.gz$/)
				|| next;
			$fname_w = "$logdir/$foo/$fname";
			((-s $fname_w) <= $mindumpsize) && next;
			$logs{$fname} = $fname_w;
			$dates{$fname} = "$foo " . localtime($1) . " to " .
				localtime($2);
		}
		closedir(DIR);
	}
}

sub do_dump
{
	my($command, $foo, $bar);
	my($line, $resp, $c);
	&get_logs();
	foreach $foo (@selected_logs)
	{
		$bar .= "$logs{$foo} ";
	}
	print "<h1 align=center>Query Results</h1><p>\n";
	print "dumps:  $bar<br>\n";
	$command = $zcat . " -f " . $bar . " | " .
		$tcpdump . " -r - -n -l -s 65536 ";
	@Verbose && ($command .= "-x ");
	$command .= join(' ', @options) . " " . join(' ', @filter) . "|";
	open(PIPE, "$command") ||
		((print "<h3>Couldn't open $command pipe</h3>\n"), return);
	print "<pre><br>\n";
	while(<PIPE>)
	{
		@Verbose || (print, next);
		/^(\s*)([0-9a-f]+\s*)+[\n\r]+$/ || (print, next);
		chomp($line = $_);
		$resp = "\t";
		while($line =~ /([0-9a-f][0-9a-f])(\s*)/)
		{
			$line = $';
			$c = chr(hex $1);
			($c =~ /[\040-\176]/) || ($c = '.');
			$resp .= $c;
		}
		print "$resp\n";
	}
	print "<br></pre><br>\n";
	close(PIPE);
}

sub do_widget
{
	my($widget) = @_;
	my($command, $foo, $bar);
	&get_logs();
	foreach $foo (@selected_logs)
	{
		$bar .= "$logs{$foo} ";
	}
	print "<h1 align=center>Query Results</h1><p>\n";
	$command = "$INSTALL_PREFIX/$BINDIR/$widget -r $bar ";
	@Verbose || ($command .= "-c ");
	$command .= join(' ', @options) . " " . join(' ', @filter) . "|";
	#print "$command\n";
	open(PIPE, "$command") ||
		((print "<h3>Couldn't open $command pipe</h3>\n"), return);
	print "<pre><br>\n";
	while(<PIPE>)
	{
		print;
	}
	print "<br></pre><br>\n";
	close(PIPE);
}

sub do_plot
{
	my($widget) = @_;
	my($command, $foo, $bar, $name);
	$name = "plot.$$";
	&get_logs();
	foreach $foo (@selected_logs)
	{
		$bar .= "$logs{$foo} ";
	}
	print "<h1 align=center>Query Results</h1><p>\n";
	$command = "$INSTALL_PREFIX/$BINDIR/plot_single.pl -r \"$bar\" -n $name";
	$command .= join(' ', @options) . " " . join(' ', @filter) . "|";
	#print "$command\n";
	open(PIPE, "$command") ||
		((print "<h3>Couldn't open $command pipe</h3><p>\n"), return);
	print "<pre><br>\n";
	while(<PIPE>)
	{
		print;
	}
	print "<br></pre><br>\n";
	close(PIPE);
	system("cp $plotdir/$name.png $htmldir/$name.png");
	print "<image src=$htmlref/$name.png border=0 alt=\"plot\">\n";
	#open(FILE, "<$plotdir/$name.png");
	#while(<FILE>)
	#{
		#print;
	#}
	#close(FILE);
	(-f "$plotdir/$name.png") && unlink("$plotdir/$name.png");
	#(-f "$htmldir/$name.png") && unlink("$htmldir/$name.png");
}

sub do_extract
{
	undef @expression;
	print "<h1>Enter an extraction expression</h1><p>\n";
	print $query->textfield('expression');
	print $query->submit('action', 'Extract');
}

sub do_xq
{
	my($foo);
	foreach $foo (keys %src)
	{
		undef @{"times_$foo"};
		undef %{"windows_$foo"};
	}
	undef %src;
	$foo = join(' ', @expression);
	$foo =~ /(\d{9,})\.(\d{6})(\s+)(\d{1,3})\.(\d{1,3})\.(\d{1,3})\.(\d{1,3})(:(\d+))?(\s+)->(\s+)(\d{1,3})\.(\d{1,3})\.(\d{1,3})\.(\d{1,3})(:(\d+))?(\s+)(\d+)(\s*)$/ || return;
	@{"times_$4.$5.$6.$7"} = ();
	@{"windows_$4.$5.$6.$7"} = ();
	push(@{"times_$4.$5.$6.$7"}, $1);
	$src{"$4.$5.$6.$7"}++;
	&figure_ranges();
	&extract_dumps();
	&get_dumps("$4.$5.$6.$7");
	&do_logs("$4.$5.$6.$7");
}

sub get_dumps
{
	my($host) = @_;
	my($fname, $fname_w, $foo);
	undef %logs;
	undef %dates;
	opendir(DIR, "$dumpdir/$host") ||
		((print "<h3>Couldn't open dir $dumpdir/$host</h3><p>\n"),
		return);
	while($fname = readdir(DIR))
	{
		($fname =~ /^$host\.(\d+)-(\d+)\.(.+)/) || next;
		$fname_w = "$dumpdir/$host/$fname";
		((-s $fname_w) <= $mindumpsize) && next;
		$logs{$fname} = $fname_w;
		$dates{$fname} = "$host " . localtime($1) . " to " .
			localtime($2) . " from $3";
	}
	closedir(DIR);
}

sub figure_ranges
{
	my($foo, $bar, $baz, $stmp, $etmp, $start, $stop, $found);
	foreach $foo (keys %src)
	{
		foreach $bar (@{"times_$foo"})
		{
			$stmp = $bar - $window;
			$etmp = $bar + $window;
			$found = 0;
			foreach $baz (keys %{"windows_$foo"})
			{
				($start, $stop) = split(':', $baz);
				($etmp >= $start) || next;
				($stmp <= $stop) || next;
				$found++;
				delete ${"windows_$foo"}{$baz};
				($stmp <= $start) && ($start = $stmp);
				($etmp >= $stop) && ($stop = $etmp);
				${"windows_$foo"}{"$start:$stop"}++;
			}
			$found && next;
			${"windows_$foo"}{"$stmp:$etmp"}++;
		}
	}
}

sub extract_dumps
{
	my($foo, $fname, $fname_w, $shost, $found, $start, $stop, $stmp, $etmp);
	foreach $foo (@log_sensors)
	{
		(-d "$logdir/$foo") || next;
		opendir(DIR, "$logdir/$foo") || next;
		while($fname = readdir(DIR))
		{
			($fname =~ /^$foo-sensor_dump\.(\d+)-(\d+)\.gz$/) ||
				next;
			$stmp = $1, $etmp = $2;
			foreach $shost (keys %src)
			{
				foreach $bar (sort(keys %{"windows_$shost"}))
				{
					($start, $stop) = split(':', $bar);
					(int($stop) >= int($stmp)) || next;
					(int($start) <= int($etmp)) || next;
					#print "<b>" . int($start) . " - " . int($stop) . "</b><br>\n";
					#print "<b>" . int($stmp) . " - " . int($etmp) . "</b><hr>\n";
					(-d "$dumpdir/$shost") ||
						mkdir("$dumpdir/$shost", 0700);
					$command = "$INSTALL_PREFIX/$BINDIR/parser -B -P -R -r $logdir/$foo/$fname -w $dumpdir/$shost/$shost.$stmp-$etmp.$foo host $shost |";
					open(PIPE, "$command") ||
						((print "<h3>Error opending $command pipe</h3><p>\n"), next);
					while(<PIPE>)
					{
						print;
					}
					close(PIPE);
				}
			}
		}
	}
}

sub check_arg
{
	my($foo) = @_;
	$foo =~ tr/[-_\/.a-zA-Z0-9]//cd;
	return($foo);
}

sub check_filter
{
	my($foo) = @_;
	$foo =~ tr/[&\[\]\(\)%!-_\ \/.a-zA-Z0-9]//cd;
	return($foo);
}
