#include "shoki_sensor.h"
#include "shoki_utils.h"
#include "shoki_common_utils.h"
#include "shoki_config.h"
#include "shoki.h"
#include "shoki_init.h"

pcap_handler out = pcap_dump_pt;
u_char *pcap_userdata = 0;
int malsize, fd;
int snaplen = DEFAULT_SNAPLEN;

char *read_file(char *fname);

int main(int argc, char **argv)
{
	register int op;
	register char *cfile = NULL;
	register char *dumplist = NULL;
	register char *filter = NULL;
	bpf_u_int32 localnet, netmask;
	struct bpf_program fcode;
	char errbuf[PCAP_ERRBUF_SIZE];
	struct sigaction *changesig;
	struct sigaction *cleansig;
	struct sigaction *childsig;
	int dflag = 0;

	shoki_type = strdup(SHOKI_TYPE);

	while((op = getopt(argc, argv, "c:C:dF:f:i:s:w:")) != EOF)
		switch(op)
		{
			case 'c':
				count = atoi(optarg);
				break;
			case 'C':
				cfile = optarg;
				break;
			case 'd':
				dflag++;
				break;
			case 'F':
				fname = optarg;
				break;
			case 'f':
				currentfile = optarg;
				break;
			case 'i':
				dev = optarg;
				break;
			case 's':
				default_snaplen = atoi(optarg);
				break;
			case 'w':
				outfile = optarg;
				break;
		}

	cleansig = malloc(sizeof(struct sigaction));
	changesig = malloc(sizeof(struct sigaction));
	childsig = malloc(sizeof(struct sigaction));

	childsig->sa_handler = (void *)clean_kids;
	childsig->sa_flags = SA_RESTART & SA_NOCLDSTOP;
	cleansig->sa_handler = (void *)cleanup;
	cleansig->sa_flags = SA_RESTART & SA_NOCLDSTOP;
	changesig->sa_handler = (void *)change_dump;
	changesig->sa_flags = SA_RESTART & SA_NOCLDSTOP;

	(void)sigaction(SIGCHLD, childsig, NULL);
	(void)sigaction(SIGTERM, cleansig, NULL);
	(void)sigaction(SIGINT, cleansig, NULL);
	(void)sigaction(SIGHUP, changesig, NULL);

	((cfile == NULL) ? (parse_shoki_config(SENSOR_CONFIG)) : 
		(parse_shoki_config(cfile)));
	if(fname) (filter = read_file(fname));
	if(filter == NULL) error("Specify a filter, dammit.");
	if(outfile == NULL) error("Specify an output file, dammit.");
	if(dev == NULL) error("Specify a device, dammit.");
	if(default_snaplen == -1) (default_snaplen = DEFAULT_SNAPLEN);
	(void)openlog(SHOKI_LOG_NAME, LOG_PID, SHOKI_LOG_SVC);
       	syslog(SHOKI_LOG_LVL, "sensor starting");
       	(void)closelog();
	if(!(dflag)) 
		if(fork()) exit(0);
	pd = pcap_open_live(dev, snaplen, 1, 1000, errbuf);
	if(pd == NULL) error("%s", errbuf);
	localnet = 0;
	netmask = 0;
	if(pcap_lookupnet(dev, &localnet, &netmask, errbuf) < 0)
		((localnet = 0), (netmask = 0),
		error("pcap_lookupnet:  %s", errbuf));
	if(pcap_compile(pd, &fcode, filter, 1, netmask) < 0) 
		error("pcap_compile (%s):  %s\n", filter, pcap_geterr(pd));
	if(pcap_setfilter(pd, &fcode) < 0) 
		error("pcap_setfilter:  %s\n", pcap_geterr(pd));
	malsize = strlen(outfile) + sizeof("123456780") + 2;
	currentdump = malloc(malsize);
	dumplist = malloc(malsize);
	if(currentdump == NULL) error("Out of memory");
	for(;;)
	{
		(void)open_dump();
		(void)capture_loop();
		(void)close_dump();
	}
	pcap_close(pd);
}

int
cleanup(int signo)
{
	if(pd != NULL && pcap_file(pd) == NULL)
	{
		(void)fflush(stdout);
		putc('\n', stderr);
	}
	(void)change_log();
	(void)openlog(SHOKI_LOG_NAME, LOG_PID, SHOKI_LOG_SVC);
       	syslog(SHOKI_LOG_LVL, "sensor shutting down");
       	(void)closelog();
	exit(0);
}

void
change_log()
{
	register char *endpoint;
	register int i;

	endpoint = malloc(512);
	i = time(NULL);
	(void)snprintf(endpoint, 512, "/bin/mv %s %s-%010d\n", currentdump, currentdump, i);
	system(endpoint);
	free(endpoint);
}

void
open_dump()
{
	register int i = 0;

	i = time(NULL);
	(void)snprintf(currentdump, malsize, "%s.%010d", outfile, i);
	if((fd = open(currentfile, O_WRONLY|O_APPEND|O_CREAT, S_IRUSR|S_IWUSR)) == -1)
		error("open:  error opening currentfile\n");
	write(fd, currentdump, strlen(currentdump));
	close(fd);
	dumper = pcap_dump_open(pd, currentdump);
	if(dumper == NULL) error("%s", pcap_geterr(pd));
	pcap_userdata = (u_char *)dumper;
}

void
capture_loop()
{
	if(pcap_loop(pd, count, out, pcap_userdata) < 0)
	error("%s", pcap_geterr(pd));
}

void
close_dump()
{
	pcap_dump_close(dumper);
	(void)change_log();
}

void
change_dump()
{
	(void)openlog(SHOKI_LOG_NAME, LOG_PID, SHOKI_LOG_SVC);
       	syslog(SHOKI_LOG_LVL, "rotating dumpfile ");
       	(void)closelog();
	pcap_dump_close(dumper);
	(void)change_log();
	(void)open_dump();
}

void
clean_kids()
{
	waitpid(-1, NULL, WNOHANG);
}

void
pcap_dump_pt(u_char *user, const struct pcap_pkthdr *h, const u_char *sp)
{
	register FILE *f;

	f = (FILE *)user;
	(void)fwrite((char *)h, sizeof(*h), 1, f);
	(void)fwrite((char *)sp, h->caplen, 1, f);
	return;
}

void
pcap_print_pt(u_char *foo, const struct pcap_pkthdr *bar, const u_char *baz)
{
}
