/*
	Not for production use.		-spb
*/
#include "shoki_includes.h"
#include "shoki.h"
#include "shoki_init.h"
#include "shoki_config.h"
#include "shoki_doctrine.h"
#include "shoki_callback.h"
#include "shoki_parser.h"
#include "shoki_server.h"

char *dumpfile = NULL;
int firstsec = 0, lastsec = 0, silent = 0, callbacks = 1;
int linux_kludge = 0, timeval_kludge = 0, fragre = 0;
struct bpf_program base_fcode;
FILE *logfile;

extern int nfilters;

pcap_handler out;
u_char *pcap_userdata = 0;

char *read_file(char *fname);

char *filter, *nonce, *hash, *input_filter;

static struct { const char *name; int option; } inputs[] =
{
	{ "nonce",	INPUT_NONCE },
	{ "hash",	INPUT_HASH },
	{ "start",	INPUT_START },
	{ "stop",	INPUT_STOP },
	{ "filter",	INPUT_FILTER },
	{ "",		0 }
};

int main(int argc, char **argv)
{
	register int op;
	register char *cfile = NULL;
	struct sigaction *cleansig;
	struct sigaction *childsig;

	shoki_type = strdup(SHOKI_TYPE);
	shoki_log_name = strdup(SHOKI_LOG_NAME);

	while((op = getopt(argc, argv, "c:C:E:F:f:i:r:s:S:")) != EOF)
		switch(op)
		{
			case 'c':
				count = atoi(optarg);
				break;
			case 'C':
				cfile = optarg;
				break;
			case 'E':
				stop = atoi(optarg);
				break;
			case 'F':
				filterlist = optarg;
				break;
			case 'f':
				currentfile = optarg;
				break;
			case 'i':
				dev = optarg;
				break;
			case 'r':
				dumpfile = optarg;
				break;
			case 's':
				default_snaplen = atoi(optarg);
				break;
			case 'S':
				start = atoi(optarg);
				break;
		}

	base_filter = get_filter(&argv[optind]);

	cleansig = malloc(sizeof(struct sigaction));
	childsig = malloc(sizeof(struct sigaction));

	childsig->sa_handler = (void *)clean_kids;
	childsig->sa_flags = SA_RESTART & SA_NOCLDSTOP;
	cleansig->sa_handler = (void *)cleanup;
	cleansig->sa_flags = SA_RESTART & SA_NOCLDSTOP;

	(void)sigaction(SIGCHLD, childsig, NULL);
	(void)sigaction(SIGTERM, cleansig, NULL);
	(void)sigaction(SIGINT, cleansig, NULL);

	((cfile == NULL) ? (parse_shoki_config(SERVER_CONFIG)) : 
		(parse_shoki_config(cfile)));
	if((pf_outformat[0] == 0) || (pf_outdelim[0] == NULL))
		(void)default_formats(pf_outformat_d, pf_outdelim_d, pf_type_d);
	if((fname != NULL) && (base_filter == NULL))
		base_filter = read_file(fname);
	if(base_filter == NULL) error("Specify a filter, dammit.");
	if(filterlist != NULL) (void)parse_config(filterlist);
	if(default_snaplen == 0) (default_snaplen = DEFAULT_SNAPLEN);
	pd = fake_pcap(pd, 0);
	(void)openlog(SHOKI_LOG_NAME, LOG_PID, SHOKI_LOG_SVC);
       	syslog(SHOKI_LOG_LVL, "server starting");
       	(void)closelog();
	if(read_input1() != 0) (void)log_exit();
	logfile = stdout;
	fflush(NULL);
 	out = pcap_hash_pt;
	(void)read_zdumps(dumpfile, out);
	printf("HASH DONE\n");
	if(read_input2() != 0) (void)log_exit();
 	out = pcap_print_pt;
	(void)read_zdumps(dumpfile, out);
	(void)openlog(SHOKI_LOG_NAME, LOG_PID, SHOKI_LOG_SVC);
       	syslog(SHOKI_LOG_LVL, "server exiting");
       	(void)closelog();
	exit(0);
}

int
cleanup(int signo)
{
	if(pd != NULL && pcap_file(pd) == NULL)
	{
		(void)fflush(stdout);
		putc('\n', stderr);
	}
	(void)openlog(SHOKI_LOG_NAME, LOG_PID, SHOKI_LOG_SVC);
       	syslog(SHOKI_LOG_LVL, "server shutting down");
       	(void)closelog();
	exit(0);
}

void
capture_loop()
{
	if(pcap_loop(pd, count, out, pcap_userdata) < 0)
	error("%s", pcap_geterr(pd));
}

void
clean_kids()
{
	waitpid(-1, NULL, WNOHANG);
}

void
pcap_print_pt(u_char *user, const struct pcap_pkthdr *h, const u_char *sp)
{
	struct timeval *tvp;
	struct filter *filter;
	int i;
	char *out;

	(const struct timeval *)tvp = &h->ts;
	if(lastsec && (tvp->tv_sec > lastsec)) return;
	if(firstsec && (tvp->tv_sec < firstsec)) return;
	if(!(filter = (struct filter *)calloc(1, sizeof(struct filter))))
		error("calloc:  pcap_print_pt");
	for(i = 1; i <= nfilters; i++)
	{
		if(bpf_filter(fcode_pt[i]->bf_insns, (u_char *)sp, h->len, h->caplen))
		{
			if(callback_pt[i] == 0) return;
			make_filter(i, filter);
			out = packet_output((sp + sizeof(struct ether_header)), h, pf_outformat, "foo", "id");
			printf("%s\n", out);
		}
	}
	free(filter);
}

void
pcap_hash_pt(u_char *user, const struct pcap_pkthdr *h, const u_char *sp)
{
	struct timeval *tvp;
	struct filter *filter;
	u_char md5buffer[16];
	char *out;
	int i, j;

	(const struct timeval *)tvp = &h->ts;
	if(lastsec && (tvp->tv_sec > lastsec)) return;
	if(firstsec && (tvp->tv_sec < firstsec)) return;
	if(!(filter = (struct filter *)calloc(1, sizeof(struct filter))))
		error("calloc:  pcap_hash_pt");
	for(i = 1, j = 0; (i <= nfilters) && (j == 0); i++)
	{
		if(bpf_filter(fcode_pt[i]->bf_insns, (u_char *)sp, h->len, h->caplen))
		{
			if(callback_pt[i] == 0) return;
			make_filter(i, filter);
			out = packet_output((sp + sizeof(struct ether_header)), h, pf_outformat, "foo", "id");
			md5_buffer(out, strlen(out), md5buffer);
			for(i = 0; i < 16; ++i)
				printf("%02x", md5buffer[i]);
			printf("\n");
			j++;
		}
	}
	free(filter);
}

int
read_input1()
{
	char line[256], *foo = NULL;
	int lnum = 0, match, i, j;

	while(fgets(line, sizeof(line), stdin))
	{
		lnum++;
		if(lnum > 2) error("Too many config lines");
		foo = line + strspn(line, WHITE);
		if(!*foo) error("Bad option, line %d", lnum);
		foo = strtok(foo, WHITE);
		for(i=0; foo[i]; i++) foo[i] = tolower(foo[i]);
		match = -1;
		for(j = 0; inputs[j].name && (match == -1); j++)
		{
			if(strncasecmp(foo, inputs[j].name, 256) == 0)
				match = inputs[j].option;
		}
		if(match == -1) error("Bad config option, line %d", lnum);
		foo = strtok(NULL, WHITE);
		if(foo == NULL) error("No argument for option, line %d", lnum);
		switch(match)
		{
			case INPUT_START:
				if(firstsec) error("start multiply defined, line %d", lnum);
				firstsec = strtol(foo, (char **)NULL, 10);
				break;
			case INPUT_STOP:
				if(lastsec) error("stop multiply defined, line %d", lnum);
				lastsec = strtol(foo, (char **)NULL, 10);
				break;
			default:
				error("Bad option, line %d", lnum);
				break;
		}
	}
	if(lastsec < firstsec) error("stop needs to be later than start");
	if((lastsec - firstsec) > query_window) error("Invalid query window");
	if(start && (firstsec < start)) error("Invalid start time");
	if(stop && (lastsec > stop)) error("Invalid stop time");
	return(0);
}

int
read_input2()
{
	return(0);
}

int
parse_input(char *foo)
{
	u_char md5buffer[16], md5buffer2[16];
	char *bar = NULL;
	int i;

	md5_buffer(bar, strlen(bar), md5buffer2);
	
	for(i = 0; i < 16; ++i)
	{
		if(md5buffer[i] != md5buffer2[i]) return(1);
	}
	return(0);
}

void
log_exit()
{
	(void)openlog(SHOKI_LOG_NAME, LOG_PID, SHOKI_LOG_SVC);
       	syslog(SHOKI_LOG_LVL, "server exiting");
       	(void)closelog();
	exit(1);
}
