#include "shoki_includes.h"
#include "shoki.h"
#include "shoki_init.h"
#include "shoki_config.h"
#include "shoki_doctrine.h"
#include "shoki_callback.h"
#include "shoki_service.h"
#include "shoki_utils.h"
#include "shoki_common_utils.h"

char *dumpfile = NULL;
int linux_kludge = 0, timeval_kludge = 0, silent = 0, callbacks = 1;
int quick = 0, fragre = 0;
struct bpf_program base_fcode;
FILE *logfile = NULL;
extern int nfilters;

struct histlist history_list;

pcap_handler out = pcap_print_pt;

#define PHASHSIZE	65537
int phash[PHASHSIZE];
#define HHASHSIZE	65537
int hhash[HHASHSIZE];
u_int32_t rhhash[HHASHSIZE];
int total = 0, minsize = -1, minperc = -1;
int phosts = 0, pother = 0, ptotal = 0, pports = 0;
int inc = 0, inc_count = 0, nextstop = 0, dflag = 0, history = 0, label = 0;

int pf_outformat[] = {};
char *pf_outdelim[] = {};
void results(int);

int
main(int argc, char **argv)
{
	register int op;
	register char *cfile = NULL, *config = NULL;

	shoki_type = strdup(SHOKI_TYPE);
	shoki_log_name = strdup(SHOKI_LOG_NAME);

	while((op = getopt(argc, argv, "C:dE:F:hHi:I:kKlL:opP:qr:s:S:tv")) != EOF)
		switch(op)
		{
			case 'C':
				cfile = optarg;
				break;
			case 'd':
				dflag++;
				break;
			case 'E':
				stop = atoi(optarg);
				break;
			case 'F':
				filterlist = optarg;
				break;
			case 'h':
				phosts = 1;
				break;
			case 'H':
				history = 1;
				break;
			case 'i':
				inc_count = atoi(optarg);
				break;
			case 'I':
				inc = atoi(optarg);
				break;
			case 'k':
				linux_kludge = 1;
				break;
			case 'K':
				timeval_kludge = 1;
				break;
			case 'l':
				label = 1;
				break;
			case 'L':
				currentfile = optarg;
				break;
			case 'o':
				pother = 1;
				break;
			case 'p':
				pports = 1;
				break;
			case 'P':
				minperc = atoi(optarg);
				break;
			case 'q':
				quick = 1;
				break;
			case 'r':
				dumpfile = optarg;
				break;
			case 's':
				default_snaplen = atoi(optarg);
				break;
			case 'S':
				start = atoi(optarg);	
				break;
			case 't':
				ptotal = 1;
				break;
			case 'v':
				verbose = 1;
				break;
		}

	base_filter = get_filter(&argv[optind]);

	if((inc_count != 0) && (inc == 0))
		inc = ((stop - start) / inc_count);
	if((inc != 0) && (start != 0)) nextstop = start + inc;
	((cfile == NULL) ? (parse_shoki_config(SERVICE_CONFIG)) :
		(parse_shoki_config(cfile)));
	if(filterlist != NULL) config = read_file(filterlist);
	if(filterlist != NULL) (void)parse_config(filterlist);
	if(dumpfile == NULL) error("No dumpfile");
	if(default_snaplen == -1) (default_snaplen = DEFAULT_SNAPLEN);
	if(currentfile != NULL)
	{
		if(*currentfile == '-')
			logfile = stdout;
		else
			logfile = fopen(currentfile, "a");
		setlinebuf(logfile);
		fflush(logfile);
	}
	(void)fake_pcap(pd, 1);
	(void)read_zdumps(dumpfile, out);
	if(nextstop == 0) (void)results(1);
	if(logfile) (void)fclose(logfile);
	exit(0);
}

void
results(int tank)
{
	int i;
	struct histlist *hc;

	if(history != 0)
	{
		for(hc = &history_list; hc->next; hc = hc->next)
		{
			log_msg(NULL, NULL, NULL, "foo:  %d", hc->index);
		}
		if(!(hc->next = (struct histlist *)calloc(1,
			sizeof(struct histlist))))
			error("calloc:  results");
		hc->index = nextstop;
		if(phosts != 0)
		{
			for(i = 0; i <= HHASHSIZE; i++)
			{
				hc->hhash[i] = hhash[i];
				hc->rhhash[i] = rhhash[i];
				hhash[i] = 0;
				rhhash[i] = 0;
			}
		}
		if(pports != 0)
		{
			for(i = 0; i <= PHASHSIZE; i++)
			{
				hc->phash[i] = phash[i];
				phash[i] = 0;
			}
		}
		if((tank != 0) || (nextstop >= (stop - inc - 1)))
			for(hc = &history_list; hc->next; hc = hc->next)
				log_msg(NULL, NULL, NULL, "%d", hc->index);
	}
	else
	{
		if(phosts != 0)
			(void)print_hosts();
		if(pports != 0)
			(void)print_ports();
	}
	if(tank != 0) exit(0);
}

void
print_hosts()
{
	int i, other = 0;
	float j;

	for(i = 0; i <= HHASHSIZE; i++)
	{
		if(hhash[i] != 0)
		{
			j = ((float)hhash[i] / (float)total) * 100.0;
			if(j > (float)minperc)
			{
				if(minperc >= 0)
					log_msg(NULL, NULL, NULL,"%15s     %10d     %6.2f", getname((u_char *)&rhhash[i]), hhash[i], j);
				else
					log_msg(NULL, NULL, NULL, "%15s     %10d", getname((u_char *)&rhhash[i]), hhash[i]);
			}
			else
				other += hhash[i];
			hhash[i] = 0;
		}
	}
	if((other != 0) && (pother != 0))
	{
		j = ((float)other / (float)total) * 100.0;
		if(minperc >= 0)
			log_msg(NULL, NULL, NULL, "%15s     %10d     %6.2f", "other", other, j);
		else
			log_msg(NULL, NULL, NULL, "%15s     %10d", "other", other);
	}
	if(ptotal != 0)
	{
		if(minperc >= 0)
			log_msg(NULL, NULL, NULL, "%15s     %10d     %6.2f", "total", total, 100.0);
		else
			log_msg(NULL, NULL, NULL, "%15s     %10d", "total", total);
	}
	other = 0;
	total = 0;
}

void
print_ports()
{
	int i, other = 0;
	float j;

	for(i = 0; i <= PHASHSIZE; i++)
	{
		if(phash[i] != 0)
		{
			j = ((float)phash[i] / (float)total) * 100.0;
			if(j > (float)minperc)
			{
				if(minperc >= 0)
					log_msg(NULL, NULL, NULL, "%5d     %10d     %6.2f", i,
						phash[i], j);
				else
					log_msg(NULL, NULL, NULL, "%5d     %10d", i, phash[i]);
			}
			else
				other += phash[i];
			phash[i] = 0;
		}
	}
	if((other != 0) && (pother != 0))
	{
		j = ((float)other / (float)total) * 100.0;
		if(minperc >= 0)
			log_msg(NULL, NULL, NULL, "other     %10d     %6.2f", other, j);
		else
			log_msg(NULL, NULL, NULL, "other     %10d", other);
	}
	if(ptotal != 0)
	{
		if(minperc >= 0)
			log_msg(NULL, NULL, NULL, "total     %10d     %6.2f", total, 100.0);
		else
			log_msg(NULL, NULL, NULL, "total     %10d", total);
	}
	other = 0;
	total = 0;
}


void
parse_header(const u_char *sp, int i, int len, const struct pcap_pkthdr *h)
{
	struct ether_header *ep;
	u_short ether_type;
	u_char *cp;
	struct ip *ip;
	u_int hlen, l, off;


	ep = calloc(1, sizeof(struct ether_header));
	memcpy(ep, sp, sizeof(struct ether_header));
	sp += sizeof(struct ether_header);
	ether_type = ntohs(ep->ether_type);
	if(ether_type != ETHERTYPE_IP)
		log_msg(NULL, NULL, NULL, "Non-IP packet");
	ip = (struct ip *)calloc(1, sizeof(struct ip));
	memcpy(ip, sp, sizeof(struct ip));
	hlen = ip->ip_hl * 4;
	l = ntohs(ip->ip_len);
	l -= hlen;
	off = ntohs(ip->ip_off);
	if((off & 0x1fff) == 0)
	{
		switch(ip->ip_p)
		{
			case IPPROTO_TCP:
				cp = (u_char *)calloc(1, sizeof(struct tcphdr));
				memcpy(cp, (sp + sizeof(struct ip) + hlen - 20), sizeof(struct tcphdr));
				parse_packet((u_char *)cp, l, (const u_char *)ip, ip->ip_p, &h->ts, i);
				free(cp);
				break;
			case IPPROTO_UDP:
				cp = (u_char *)calloc(1, sizeof(struct udphdr));
				memcpy(cp, (sp + sizeof(struct ip) + hlen - 20), sizeof(struct udphdr));
				parse_packet((u_char *)cp, l, (const u_char *)ip, ip->ip_p, &h->ts, i);
				break;
				free(cp);
			case IPPROTO_ICMP:
				break;
			default:
				break;
		}
	}
	free(ep);
	free(ip);
}

void
parse_packet(register const u_char *bp, register u_int length, register const u_char *bp2, register u_char proto, const struct timeval *tvp, int filt)
{
	struct tcphdr *tp;
	struct udphdr *up;
	struct ip *ip;
	u_short sport = 0, dport = 0;
	int i = 0, j = 0;
	u_int32_t ind = 0;

	ip = (struct ip *)calloc(1, sizeof(struct ip));
	memcpy(ip, bp2, sizeof(struct ip));
	switch(ip->ip_p)
	{
		case IPPROTO_TCP:
			tp = (struct tcphdr *)calloc(1, sizeof(struct tcphdr));
			memcpy(tp, bp, sizeof(struct tcphdr));
			sport = ntohs(tp->th_sport);
			dport = ntohs(tp->th_dport);
			free(tp);
			break;
		case IPPROTO_UDP:
			up = (struct udphdr *)calloc(1, sizeof(struct udphdr));
			memcpy(up, bp, sizeof(struct udphdr));
			sport = ntohs(up->uh_sport);
			dport = ntohs(up->uh_dport);
			free(up);
			break;
	}
	if(pports != 0)
	{
		if(dflag)
			phash[dport] += ntohs(ip->ip_len);
		else
			phash[sport] += ntohs(ip->ip_len);
	}
	if(phosts != 0)
	{
		if(dflag)
			ind = ip->ip_dst.s_addr;
		else
			ind = ip->ip_src.s_addr;
		i = ind % 65537;
		while((rhhash[i] != 0) && (rhhash[i] != ind))
		{
			i++;
			j++;
			if(j >= 65536)
				error("Too damn many hash collisions.  Giving up in disgust.");
		}
		hhash[i] += ntohs(ip->ip_len);
		rhhash[i] = ind;
	}
	total += ntohs(ip->ip_len);
	free(ip);
}

void
pcap_print_pt(u_char *user, const struct pcap_pkthdr *h, const u_char *sp)
{
	int foo, i;

	if((start == 0) && ((inc_count != 0) || (inc != 0)))
	{
		start = h->ts.tv_sec;
		nextstop = start + inc;
	}
	if(start != 0) if(h->ts.tv_sec < start) return;
	if(stop != 0) if((h->ts.tv_sec > stop) && (quick != 0))
		(void)results(1);
	if(stop != 0) if(h->ts.tv_sec > stop) return;
	foo = 0;
	if(nfilters == 0) error("No filters defined");
	for(i = 1; i <= nfilters && foo == 0; i++)
	{
		if(bpf_filter(fcode_pt[i]->bf_insns, (u_char *)sp, h->len, h->caplen))
		{
			foo = 1;
			parse_header(sp, i, h->len, h);
		}
	}
	if(nextstop != 0)
		if(h->ts.tv_sec > nextstop)
		{
			if(label)
				log_msg(NULL, NULL, NULL, "%d:%d",
					(nextstop - inc), nextstop);
			nextstop += inc;
			(void)results(0);
		}
}
