#include "shoki_includes.h"
#include "shoki.h"
#include "shoki_init.h"
#include "shoki_config.h"
#include "shoki_doctrine.h"
#include "shoki_callback.h"
#include "shoki_session.h"
#include "shoki_utils.h"
#include "shoki_common_utils.h"

char *dumpfile = NULL;
int callbacks = 1, linux_kludge = 0, timeval_kludge = 0;
int silent = 0, fragre = 0;
struct bpf_program base_fcode;
FILE *logfile = NULL;
extern int nfilters;

pcap_handler out = pcap_print_pt;

static struct thash thash[THASHSIZE1][THASHSIZE2];


int
main(int argc, char **argv)
{
	register int op;
	register char *cfile = NULL, *config = NULL;
	struct sigaction *cleansig, *childsig;
	int dflag = 0;

	shoki_type = strdup(SHOKI_TYPE);
	shoki_log_name = strdup(SHOKI_LOG_NAME);

	count = 0;
	while((op = getopt(argc, argv, "BcC:dE:F:kKr:Rs:S:T")) != EOF)
		switch(op)
		{
			case 'B':
				callbacks = 0;
				break;
			case 'c':
				count = 1;
				break;
			case 'C':
				cfile = optarg;
				break;
			case 'd':
				dflag = 1;
				break;
			case 'E':
				stop = atoi(optarg);
				break;
			case 'F':
				filterlist = optarg;
				break;
			case 'k':
				linux_kludge = 1;
				break;
			case 'K':
				timeval_kludge = 1;
				break;
			case 'r':
				dumpfile = optarg;
				break;
			case 's':
				default_snaplen = atoi(optarg);
				break;
			case 'S':
				start = atoi(optarg);	
				break;
		}


	base_filter = get_filter(&argv[optind]);

	cleansig = malloc(sizeof(struct sigaction));
	childsig = malloc(sizeof(struct sigaction));
	
	childsig->sa_handler = (void *)clean_kids;
	childsig->sa_flags = SA_RESTART & SA_NOCLDSTOP;
	cleansig->sa_handler = (void *)cleanup;
	cleansig->sa_flags = SA_RESTART & SA_NOCLDSTOP;

	(void)sigaction(SIGCHLD, childsig, NULL);
	(void)sigaction(SIGTERM, cleansig, NULL);
	(void)sigaction(SIGINT, cleansig, NULL);

	((cfile == NULL) ? (parse_shoki_config(SESSION_CONFIG)) :
		(parse_shoki_config(cfile)));
	if((pf_outformat[0] == 0) || (pf_outdelim[0] == NULL))
		(void)default_formats(pf_outformat_d, pf_outdelim_d, pf_type_d);
	if(filterlist != NULL) config = read_file(filterlist);
	if(dumpfile == NULL) error("No dumpfile");
	if(filterlist != NULL) (void)parse_config(filterlist);
	if(default_snaplen == -1) (default_snaplen = DEFAULT_SNAPLEN);
	pd = fake_pcap(pd, 1);
	(void)read_zdumps(dumpfile, out);
	if(count == 0)
		(void)print_conv();
	else
		(void)print_count();
	exit(0);
}

void
print_conv()
{
	struct thash *t;
	int i, j, k = 0;
	struct ip *ip;
	struct tcphdr *tp;
	struct timeval *tvp;
	struct pcap_pkthdr *h;
	char *out, *sp;
	char foo[20];

	if(!(ip = (struct ip *)calloc(1, sizeof(struct ip))))
		error("calloc:  print_conv");
	if(!(tp = (struct tcphdr *)calloc(1, sizeof(struct tcphdr))))
		error("calloc:  print_conv");
	if(!(tvp = (struct timeval *)calloc(1, sizeof(struct timeval))))
		error("calloc:  print_conv");
	if(!(sp = (char *)calloc(1, (sizeof(struct ip) + sizeof(struct tcphdr)))))
		error("calloc:  print_conv");
	if(!(h = (struct pcap_pkthdr *)calloc(1, sizeof(struct pcap_pkthdr))))
		error("calloc:  print_conv");
	for(i = 0; i < THASHSIZE1; i++)
	{
		for(j = 0; j < THASHSIZE2; j++)
		{
			for(t = &thash[i][j]; t && t->next; t = t->next)
			{
				if(t->ip == NULL) break;
				if(t->tp == NULL) break;
				if(t->tvp == NULL) break;
				memcpy(ip, t->ip, sizeof(struct ip));
				memcpy(tp, t->tp, sizeof(struct tcphdr));
				memcpy(tvp, t->tvp, sizeof(struct timeval));
				memcpy(&h->ts, t->tvp, sizeof(struct timeval));
				memcpy(sp, (char *)ip, sizeof(struct ip));
				memcpy((char *)(sp + sizeof(struct ip)), (char *)tp, sizeof(struct tcphdr));
				sprintf(foo, "session %d", k);
				out = packet_output(sp, h, pf_outformat, foo, NULL, sp);
				printf("%s\n", out);
			}
			if((t = &thash[i][j]) && t->next)
			{
				printf("\n");
				k = (k + 1) % 10000000000;
			}
		}
	}
	free(ip);
	free(tp);
	free(tvp);
	free(h);
	free(sp);
}

void
print_count()
{
	struct thash *t;
	int i, j, s_len = 0, d_len = 0, foo = 0;
	struct ip *ip;
	struct tcphdr *tp;
	struct timeval *tvp;
	char *sp;
	u_long s_addr = 0, d_addr;
	u_char *saddr, *daddr;
	u_short sport = 0, dport = 0;

	if(!(ip = (struct ip *)calloc(1, sizeof(struct ip))))
		error("calloc:  print_count");
	if(!(tp = (struct tcphdr *)calloc(1, sizeof(struct tcphdr))))
		error("calloc:  print_count");
	if(!(tvp = (struct timeval *)calloc(1, sizeof(struct timeval))))
		error("calloc:  print_count");
	if(!(sp = (char *)calloc(1, (sizeof(struct ip) + sizeof(struct tcphdr)))))
		error("calloc:  print_count");
	if(!(saddr = (u_char *)calloc(1, sizeof(struct in_addr))))
		error("calloc:  print_count");
	if(!(daddr = (u_char *)calloc(1, sizeof(struct in_addr))))
		error("calloc:  print_count");
	for(i = 0; i < THASHSIZE1; i++)
	{
		for(j = 0; j < THASHSIZE2; j++)
		{
			s_len = 0;
			d_len = 0;
			s_addr = 0;
			d_addr = 0;
			foo = 0;
			for(t = &thash[i][j]; t && t->next; t = t->next)
			{
				if(t->ip == NULL) break;
				if(t->tp == NULL) break;
				if(t->tvp == NULL) break;
				foo++;
				memcpy(ip, t->ip, sizeof(struct ip));
				if((s_addr == 0) && (d_addr == 0))
				{
					s_addr = ip->ip_src.s_addr;
					d_addr = ip->ip_dst.s_addr;
					memcpy(saddr, &ip->ip_src, sizeof(struct in_addr));
					memcpy(daddr, &ip->ip_dst, sizeof(struct in_addr));
					memcpy(tp, t->tp, sizeof(struct tcphdr));
					sport = ntohs(tp->th_sport);
					dport = ntohs(tp->th_dport);
				}
				if((s_addr == ip->ip_src.s_addr) &&
					(d_addr == ip->ip_dst.s_addr))
				{
					s_len += ntohs(ip->ip_len);
				}
				else
				if((d_addr == ip->ip_src.s_addr) &&
					(s_addr == ip->ip_dst.s_addr))
				{
					d_len += ntohs(ip->ip_len);
				}
				else
				{
					printf("Oddball:  %s -> ", ipaddr_string(&ip->ip_src));
					printf("%s\n", ipaddr_string(&ip->ip_dst));
				}
			}
			if(foo != 0)
			{
				printf("%s:%d (%d) -> ", ipaddr_string(saddr), sport, s_len);
				printf("%s:%d (%d)\n", ipaddr_string(daddr), dport, d_len);
			}
		}
	}
	free(ip);
	free(tp);
	free(tvp);
	free(sp);
}


void
cleanup(int signo)
{
	exit(0);
}

void
clean_kids()
{
	waitpid(-1, NULL, WNOHANG);
}

void
parse_header(const u_char *sp, int i, int len, const struct pcap_pkthdr *h)
{
	struct ether_header *ep;
	u_short ether_type;
	u_char *cp;
	struct ip *ip;
	u_int hlen, l, off;


	if(!(ep = calloc(1, sizeof(struct ether_header))))
		error("calloc:  parse_header");
	memcpy(ep, sp, sizeof(struct ether_header));
	sp += sizeof(struct ether_header);
	ether_type = ntohs(ep->ether_type);
	if(ether_type != ETHERTYPE_IP)
		printf("Non-IP packet\n");
	if(!(ip = (struct ip *)calloc(1, sizeof(struct ip))))
		error("calloc:  parse_header");
	memcpy(ip, sp, sizeof(struct ip));
	if(!(cp = (u_char *)calloc(1, sizeof(struct tcphdr))))
		error("calloc:  parse_header");
	hlen = ip->ip_hl * 4;
	l = ntohs(ip->ip_len);
	l -= hlen;
	off = ntohs(ip->ip_off);
	if((off & 0x1fff) == 0)
	{
		memcpy(cp, (sp + sizeof(struct ip) + hlen - 20), sizeof(struct tcphdr));
		switch(ip->ip_p)
		{
			case IPPROTO_TCP:
				parse_packet((u_char *)cp, l, (const u_char *)ip, ip->ip_p, &h->ts, i);
				break;
			case IPPROTO_UDP:
			case IPPROTO_ICMP:
				break;
			default:
				break;
		}
	}
	free(cp);
	free(ep);
	free(ip);
}

void
parse_packet(register const u_char *bp, register u_int length, register const u_char *bp2, register u_char proto, const struct timeval *tvp, int filt)
{
	struct tcphdr *tp;
	struct ip *ip, *ip2;
	register u_int hlen = 0;
	u_short sport = 0, dport = 0;
	u_short sec, usec;
	u_int32_t seq = 0, ack = 0, sa1 = 0, sa2 = 0, sas = 0, sad = 0;
	u_int64_t ind = 0;
	struct tconv tc;
	struct thash *t;
	int dir, i;

	if(!(ip = (struct ip *)calloc(1, sizeof(struct ip))))
		error("calloc:  parse_packet");
	if(!(ip2 = (struct ip *)calloc(1, sizeof(struct ip))))
		error("calloc:  parse_packet");
	if(!(tp = (struct tcphdr *)calloc(1, sizeof(struct tcphdr))))
		error("calloc:  parse_packet");
	memcpy(ip, bp2, sizeof(struct ip));
	memcpy(tp, bp, sizeof(struct tcphdr));
	sport = ntohs(tp->th_sport);
	dport = ntohs(tp->th_dport);
	hlen = tp->th_off * 4;
	length -= hlen;
	sec = (u_short)tvp->tv_sec;
	usec = (u_short)tvp->tv_usec;

	/* The logic for this hashing is stolen from the tcpdump(8) source.
								-spb
	*/
	if(sport < dport || (sport == dport && ip->ip_src.s_addr <
		ip->ip_dst.s_addr))
	{
		tc.src = ip->ip_src;
		tc.dst = ip->ip_dst;
		tc.port = (sport << 16 | dport);
		ind = (ip->ip_src.s_addr >> 24 | ip->ip_dst.s_addr);
		sa1 = ip->ip_src.s_addr;
		sa2 = ip->ip_dst.s_addr;
		dir = 0;
	}
	else
	{
		tc.src = ip->ip_dst;
		tc.dst = ip->ip_src;
		tc.port = (dport << 16 | sport);
		ind = (ip->ip_dst.s_addr >> 24 | ip->ip_src.s_addr);
		sa1 = ip->ip_dst.s_addr;
		sa2 = ip->ip_src.s_addr;
		dir = 1;
	}
	/* FOO */
	t = &thash[ind % THASHSIZE1][tc.port % THASHSIZE2];
	i = 0;
	while(t && t->ip && (ind < THASHSIZE1))
	{
		memcpy(ip2, t->ip, sizeof(struct ip));
		sas = ip2->ip_src.s_addr;
		sad = ip2->ip_src.s_addr;
		if((sas == 0) && (sad == 0)) break;
		if(((sa1 == sas) && (sa2 == sad)) ||
			((sa2 == sas) && (sa1 == sad))) break;
		ind++;
		i++;
		if(i > THASHSIZE1)
			error("Too many hash collisions.  Giving up in disgust.");
		t = &thash[ind % THASHSIZE1][tc.port % THASHSIZE2];
	}
	for(t = &thash[ind % THASHSIZE1][tc.port % THASHSIZE2]; t->next; t = t->next);
	if(!(t->tp = (char *)calloc(1, sizeof(struct tcphdr))))
		error("calloc:  parse_packet");
	if(!(t->ip = (char *)calloc(1, sizeof(struct ip))))
		error("calloc:  parse_packet");
	if(!(t->tvp = (char *)calloc(1, sizeof(struct timeval))))
		error("calloc:  parse_packet");
	memcpy(t->ip, ip, sizeof(struct ip));
	memcpy(t->tp, tp, sizeof(struct tcphdr));
	memcpy(t->tvp, tvp, sizeof(struct timeval));
	if(!(t->next = (struct thash *)calloc(1, sizeof(struct thash))))
		error("calloc:  parse_packet");
	t->addr = tc;
	if(dir)
	{
		t->ack = seq;
		t->seq = ack - 1;
	}
	else
	{
		t->seq = seq;
		t->ack = ack - 1;
	}
	free(tp);
	free(ip);
	free(ip2);
}

void
pcap_print_pt(u_char *user, const struct pcap_pkthdr *h, const u_char *sp)
{
	int foo, i;
	if(start != 0) if(h->ts.tv_sec < start) return;
	if(stop != 0) if(h->ts.tv_sec > stop) return;
	foo = 0;
	if(nfilters == 0) error("No filters defined");
	for(i = 1; i <= nfilters && foo == 0; i++)
	{
		if(bpf_filter(fcode_pt[i]->bf_insns, (u_char *)sp, h->len, h->caplen))
		{
			foo = 1;
			parse_header(sp, i, h->len, h);
		}
	}
}
